<?php
/**
 * @file
 * Drupal RestMini module.
 */


class RestMini {

  /**
   * @var integer
   */
  protected static $errorCodeOffset = 0;

  /**
   * @var array
   */
  protected static $errorCodes = array();

  /**
   * Get error code by name, or code list, or code range.
   *
   * @param string $name
   *   Non-empty: return code by name (defaults to 'unknown')
   *   Default: empty (~ return codes list).
   * @param boolean $range
   *   TRUE: return code range array(N-first, N-last).
   *   Default: FALSE (~ ignore argument).
   *
   * @return integer|array
   */
  public static function errorCode($name = '', $range = FALSE) {
    static $codes;

    if ($name) {
      return static::$errorCodeOffset
        + (array_key_exists($name, static::$errorCodes) ? static::$errorCodes[$name] : static::$errorCodes['unknown']);
    }

    if ($range) {
      return array(
        static::$errorCodeOffset,
        // Range of sub modules should only be 100, to allow for all sub modules within an overall range of 1000.
        static::$errorCodeOffset + 99
      );
    }

    if (!$codes) {
      $codes = static::$errorCodes; // Copy.
      if (($offset = static::$errorCodeOffset)) {
        foreach ($codes as &$code) {
          $code += $offset;
        }
        unset($code); // Iteration ref.
      }
    }

    return $codes;
  }


  /**
   * Must not be overridden as the property is used by methods (of RestMiniService) that are final.
   * Should be final, but PHP doesn't support final properties (only methods).
   *
   * @var array
   */
  protected static $supportedMethods = array('HEAD', 'GET', 'POST', 'PUT', 'DELETE');


  /**
   * Get status text of a status code.
   *
   * Does not take HTTP method into consideration - a 201 as response to PUT should rather be 'Updated' than 'Created'.
   *
   * @param integer $code
   *
   * @return string
   *   If unsupported status $code: 'Unknown Status'.
   */
  final public static function statusText($code) {
    // Stringify code to prevent coercion fault; a string matches any number in a PHP switch.
    switch ('' . $code) {
      case '200':
        return 'OK';
      case '201':
        return 'Created';
      case '204':
        return 'No Content';
      case '301':
        return 'Moved Permanently';
      case '302':
        return 'Found';
      case '304':
        return 'Not Modified';
      case '400':
        return 'Bad Request';
      case '401':
        return 'Unauthorized';
      case '403':
        return 'Forbidden';
      case '404':
        return 'Not Found';
      case '405':
        return 'Method Not Allowed';
      case '406':
        return 'Not Acceptable';
      case '410':
        return 'Gone';
      case '415':
        return 'Unsupported Media Type';
      case '500':
        return 'Internal Server Error';
      case '501':
        return 'Not Implemented';
      case '502':
        return 'Bad Gateway';
      case '503':
        return 'Service Unavailable';
      case '504':
        return 'Gateway Timeout';
    }
    return 'Unknown Status';
  }

  /**
   * Get info about appropriate use of HTTP response status codes to HTTP request methods.
   *
   * @param string $method
   *   Non-empty: list status codes usable by that HTTP method.
   *   Default: empty (~ list all status codes and the HTTP methods that may use them).
   *
   * @return array
   */
  final public static function statusCodes($method = '') {
    static $codes, $all_methods;

    if (!$codes) {
      $all_methods = self::$supportedMethods;
      $codes = array(
        200 => array('HEAD', 'GET', 'DELETE'), // OK.
        201 => array('POST', 'PUT'), // Created|Updated (with payload).
        204 => array('POST', 'PUT', 'DELETE'), // No Content.
        301 => $all_methods, // Moved Permanently.
        302 => $all_methods, // Found.
        304 => array('HEAD', 'GET'),
        400 => $all_methods, // Bad Request.
        401 => $all_methods, // Unauthorized.
        403 => $all_methods, // Forbidden.
        404 => $all_methods, // Not Found.
        405 => $all_methods, // Method Not Allowed.
        406 => $all_methods, // Not Acceptable.
        410 => $all_methods, // Gone.
        415 => array('POST', 'PUT'), // Unsupported Media Type.
        500 => $all_methods, // Internal Server Error.
        501 => $all_methods, // Not Implemented.
        502 => $all_methods, // Bad Gateway.
        503 => $all_methods, // Service Unavailable.
        504 => $all_methods, // Gateway Timeout.
      );
    }
    if (!$method) {
      return $codes;
    }
    $arr = array();
    if (in_array($method, $all_methods, TRUE)) {
      foreach ($codes as $status => $methods) {
        if (in_array($method, $methods)) {
          $arr[$status] = $status;
        }
      }
    }
    return $arr;
  }

  /**
   * Check if a successful status - 200, 201, 204, 304 - and that status and optionally content length matches HTTP method.
   *
   * Content length: consider that neither that neither '0' nor an empty array is an empty response.
   *
   * @param string $method
   *   HEAD|GET|POST|PUT|DELETE.
   * @param integer $status
   *   Integer: check against method.
   *   No argument: return array of successful status code of that method.
   * @param integer $contentLength
   *   Integer: check against method and status.
   *   No argument: don't consider content length.
   *
   * @return boolean|array|NULL
   *   Boolean: whether status (and optionally content length) matches method.
   *   Array: list of successful status code of that method.
   *   NULL: on error.
   */
  final public function statusSuccess($method, $status = 0, $contentLength = 0) {
    $nArgs = func_num_args();

    switch ($method) {
      case 'HEAD':
      case 'GET':
        $codes = array(200, 304);
        break;
      case 'POST':
      case 'PUT':
        $codes = array(201, 204); // @todo: W3C says 200 is fine too.
        break;
      case 'DELETE':
        $codes = array(200, 204);
        break;
      default:
        static::log(
          'restmini',
          'HTTP method not supported',
          NULL,
          $method,
          WATCHDOG_ERROR
        );
        return NULL;
    }

    if ($nArgs == 1) {
      return $codes;
    }

    if (!in_array($status, $codes)) {
      return FALSE;
    }
    if ($nArgs == 2) {
      return TRUE;
    }

    return ($method == 'HEAD' || $status == 204 || $status == 304) ? !$contentLength : !!$contentLength;
  }

  /**
   * Logs to watchdog.
   *
   * Traces Exception, if $exception and the Inspect module exists.
   *
   * @param string $type
   * @param string $message
   * @param Exception|NULL $exception
   * @param mixed $variable
   * @param integer $severity
   */
  protected static function log($type, $message, $exception = NULL, $variable = NULL, $severity = WATCHDOG_ERROR) {
    if (module_exists('inspect')) {
      if ($exception) {
        inspect_trace($exception, array(
          'category' => $type,
          'message' => $message,
          'severity' => $severity,
          'wrappers' => 1,
        ));
      }
      else {
        inspect($variable, array(
          'category' => $type,
          'message' => $message,
          'severity' => $severity,
          'wrappers' => 1,
        ));
      }
    }
    else {
      watchdog(
        $type,
        (!$message && $exception) ? $exception->getMessage() : $message,
        array(),
        $severity
      );
    }
  }

}
